/*
	Micha Plutecki G1SIS, 27.03.2008

	Wzorzec projektowy dekorator (decorator) - prezentacja

	Mechanizm dziaania:
		Opakowywuje 

	Zastosowanie:
	>	Rozszerza funkcjonalno klasy, zapobiegajc tworzeniu 
		duej liczby klas z kombinacjami rnych funkcjonalnoci.
	>	Jako alternatywa dla dziedziczenia, pozwala na 
		rozszerzanie klasy podczas dziaania aplikacji.
	>	W szczeglnoci moe zosta uyty do kontroli dostpu 
		do klasy poprzez opakowanie jej w dekorator kontrolii 
		dostpu.
*/

/*
	Zastosowane biblioteki
*/
#include <iostream>
#include <string>

/*
	Abstrakcyjna klasa bazowa stanowica interfejs dla gwnej 
	klasy programu oraz jej dekoratora.
	Dostarcza dwie czysto wirtualne metody.
*/
class ICoffee {
public:
	ICoffee() { }
	virtual ~ICoffee() { }

	virtual float getPrice() = 0;
	virtual void getDescription() = 0;
};

/*
	Gwna klasa programu, ktra moe zosta uyta np. 
	do oprogramowania automatu do kawy :]. Klasa dziedziczy 
	po abstrakcyjnej klasie bazowej ICoffee i realizuje 
	jej metody wirtualne.
*/
class Coffee : public ICoffee {
public:
	//Wypenia pola klasy Coffee, definiujc j.
	Coffee() {
		price_ = 8.0;
		description_ = "A coffee";
	}
	virtual ~Coffee() { }
	
	//Metoda zwraca aktualn cen kawy.
	float getPrice() {
		return price_;
	}

	//Metoda zwraca opis kawy.
	void getDescription() {
		std::cout << description_;
	}

private:
	//Pola klasy Coffee opisujce kaw, dostpne 
	//tylko przez powysze metody.
	float price_;
	std::string description_;
};

/*
	Klasa dekoratora, dziedziczy po abstrakcyjnej klasie 
	bazowej ICoffee.
*/
class CoffeeDecorator : public ICoffee {
public:
	//Konstruktor tworzy relacj midzy klas dekoratora 
	//i gwn klas programu (Coffee).
	CoffeeDecorator(ICoffee *c) {
		coffee = c;
	}
	//Konieczne wywoanie destruktora w klasie bazowej.
	virtual ~CoffeeDecorator() { }
	
	//Wywoanie metody getDescription() dekoratora. Umoliwia 
	//wywoanie skorelowanej metody gwnej klasy programu
	//(Coffee).
	void getDescription() {
		coffee->getDescription();
	}
	
	//Wywoanie metody getPrice() dekoratora. Umoliwia
	//wywoanie skorelowanej metody gwnej klasy aplikacji
	//(Coffee).
	float getPrice() {
		return coffee->getPrice();
	}

private:
	//Wska?nik na instancj abstrakcyjnej klasy ICoffee.
	//Umoliwia odwoanie si do gwnej klasy aplikacji
	//(Coffee).
	ICoffee *coffee;
};

/*
	Klasa pierwszego dekoratora. Dodaje mleko do kawy.
*/
class MilkCoffeeDecorator : public CoffeeDecorator {
public:
	//Konstruktor korzysta z konstruktora klasy bazowej.
	MilkCoffeeDecorator(ICoffee *c) : CoffeeDecorator(c) {}
	virtual ~MilkCoffeeDecorator() {}

	void getDescription() {
		//Zapewnia funkcjonalno klasy bazowej.
		CoffeeDecorator::getDescription();
		//Dodaje now funkcjonalno, rozszerzajc
		//opis o mleko do kawy.
		std::cout << " with milk";
	}

	float getPrice() {
		//Zapewnia funkcjonalno klasy bazowej dodajc
		//funkcjonalno wasn - dodanie ceny mleka do kawy.
		return CoffeeDecorator::getPrice() + 1.0;
	}
};

/*
	Klasa drugiego dekoratora. Dodaje cukier do kawy.
*/
class SugarCoffeeDecorator : public CoffeeDecorator {
public:
	//Konstruktor korzysta z konstruktora klasy bazowej.
	SugarCoffeeDecorator(ICoffee *c) : CoffeeDecorator(c) {}
	virtual ~SugarCoffeeDecorator() { }

	void getDescription() {
		//Zapewnia funkcjonalno klasy bazowej.
		CoffeeDecorator::getDescription();
		//Dodaje now funkcjonalno, rozszerzajc
		//opis o cukier do kawy.
		std::cout << " with sugar";
	}

	float getPrice() {
		//Zapewnia funkcjonalno klasy bazowej dodajc
		//funkcjonalno wasn - dodanie ceny cukru do kawy.
		return CoffeeDecorator::getPrice() + 0.6;
	}
};

int main(){
	//Uytkownik tworzy instancj klasy Coffee (new Coffee), korzystajc z interfejsu ICoffee.
	//Dodatkowo opakowywuje klase Coffee najpierw klas dekoratora MilkCoffeeDecorator a nastpnie
	//dwukrotnie klas dekoratora SugarCoffeeDecorator.
	ICoffee *aCoffee = new SugarCoffeeDecorator(new SugarCoffeeDecorator(new MilkCoffeeDecorator(new Coffee)));

	//Wywietlenie opisu sporzdzonej kawy.
	std::cout << "What we have got is: ";
	aCoffee->getDescription();
	std::cout << "." << std::endl;

	//Wywietlenie ceny sporzdzonej kawy.
	std::cout << "The price of the coffee is: ";
	std::cout << aCoffee->getPrice();
	std::cout << " PLN" << std::endl;

	/*	Wynik:
	What we have got is: A coffee with milk with sugar with sugar.
	The price of the coffee is: 10.2 PLN
	*/

	//Usunicie obiektu testowego.
	delete aCoffee;

	return 0;
}
